(*

The file defines the AST datatype that represents programs that can be
read in by the MiniML interpreter.  The syntax that is supported is as follows:

TopLevel ::= Decl | Expr

Decl ::= "val" Pattern "=" Expr
       | "fun" {Id Pattern [":" Type] "=" Expr}/"|"
       | "datatype" Id "=" {Id ["of" Type]}/"|"

Expr ::= Id				     (* a variable reference *)
       | Int | String | "true" | "false" | "nil"  (* literal constants *)
       | "(" ")"			     (* the unit value *)
       | "(" Expr "," Expr {"," Expr} ")"    (* a tuple of 2 or more exprs *)
       | Unop Expr | Expr Binop Expr         (* unary and binary expressions *)
       | Id Expr                             (* a function call *)
       | "if" Expr "then" Expr "else" Expr   (* an if expression *)
       | "let" {Decl} "in" Expr "end"        (* a let expression *)
       | "(" Expr ")"                        (* a simple parenthesized expression *)
       | "fn" Pattern[":" Type] "=>" Expr    (* anonymous function *)

Unop ::= "~" | "not"

Binop ::= "+" | "-" | "*" | "div" | "mod" 
        | "=" | "<>" | "<" | "<=" | ">=" | ">"
        | "^" | "::"

Pattern ::= Id [":" Type] | "_" [":" Type] (* variable and wildcard patterns *)
          | Int | String | "true" | "false"   (* literal patterns *)
          | "nil" [":" Type]	        
          | "(" ")"			(* the unit pattern *)
          | "(" Pattern "," Pattern {"," Pattern} ")"	(* a tuple pattern *)
	  | Pattern "::" Pattern	(* a list pattern *)
          | "(" Pattern ")"		(* a simple parenthesized pattern *)


Type ::= "int" | "string" | "bool" | "unit"	(* base types *)
	  | TypeVar				(* type variables *)
	  | Type {"*" Type}			(* tuple types *)
	  | Type "list"				(* list types *)
	  | Type "->" Type			(* function types *)
          | "(" Type ")"		(* a simple parenthesized type *)
	  	  	  
*)

(* Notice how nicely the ML datatype declarations parallel the grammar *)


structure Ast = struct

datatype Decl =
    ValDecl of {lhs:Pattern, rhs:Expr}
  | FunDecl of {name:string, formal:Pattern, resType:Type option, body:Expr}
                   list

and Expr =
    VarExpr of string
  | IntExpr of int
  | StringExpr of string
  | BoolExpr of bool
  | NilExpr of {typ:Type option}
  | TupleExpr of Expr list
  | UnopExpr of {operator:string, arg:Expr}
  | BinopExpr of {arg1:Expr, operator:string, arg2:Expr}
  | CallExpr of {function:string, arg:Expr}
  | IfExpr of {testExpr:Expr, thenExpr:Expr, elseExpr:Expr}
  | LetExpr of {decls:Decl list, body:Expr}
  | FnExpr of {formal:Pattern, resType:Type option, body:Expr} list

and Pattern =
    VarPattern of {name:string, typ:Type option}
  | WildcardPattern of {typ:Type option}
  | IntPattern of int
  | StringPattern of string
  | BoolPattern of bool
  | NilPattern of {typ:Type option}
  | TuplePattern of Pattern list
  | ConsPattern of {head:Pattern, tail:Pattern}

and Type =
    IntType
  | StrType
  | BoolType
  | UnitType
  | TypeVar of string
  | TupleType of Type list
  | ListType of Type
  | FunctionType of {argType:Type, resType:Type}
                    
datatype TopLevel =
    TopLevelDecl of Decl
  | TopLevelExpr of Expr

(* Convenience synonym for the argument of the FunDecl constructor. *)
type NamedFunRecord =
     {name:string, formal:Pattern, resType:Type option, body:Expr}

end
