(*

This is the main file that defines and invokes the read-eval-print
loop function.

*)

structure Repl : sig

    (* The main read-eval-print loop.  This is the function to invoke
       to start the interpreter. *)
    val repl : unit -> string

end
  =
struct

local
  open Ast
  open AstPrint
  open Parser
  open Value
  open ValuePrint
  open Env
  open Eval
in

(* a helper to print out a welcome message when starting up the
   interpreter *)
fun printWelcome() =
        (print("\n");
         print("Welcome to MiniML!\n");
	 print("Enter the input declaration or expression to be evaluated.\n");
	 print("Enter a blank line to exit.\n");
	 print("Note: this interpreter does no static typechecking!\n\n"))


(* raised when the input is empty *)
exception EmptyInput

(* a function to read in the top-level decl or expr to evaluate,
   possibly over multiple lines.  raises EmptyInput if the input was
   blank *)
fun readInput():string =
    let val initialPrompt  = "MiniML - "
	val continuePrompt = "MiniML = "
	(* a helper to read a single line of input, possibly extending
	   previous lines *)
	fun readLine(prompt:string, previousInput:string):string =
            let    
                (* print this line's prompt *)
		val _ = print(prompt);

		(* read a line of input *)
		val inputString:string = TextIO.inputLine(TextIO.stdIn)

		(* concatenate with input from previous lines *)
		val fullInputString:string = previousInput ^ inputString

		(* helper to check whether the string ends with a ";" *)
		fun ends_with_semicolon(s:string):bool =
		    let fun semicolon_helper(s:string, pos:int):bool =
			    if pos < 0 then raise EmptyInput
			    else
				let val last = String.sub(s, pos) 
				in
				    if last = #";" then true
				    else if last = #" " orelse 
					    last = #"\t" orelse 
					    last = #"\n" then
					 (* skip past whitespace char
					    and retry *)
					 semicolon_helper(s, pos-1)
		                    else false
		                end
		    in semicolon_helper(s, size(s)-1) end
	    in
		if ends_with_semicolon(fullInputString) then
		    (* we're done with input; return the full string we've
		       built so far *)
		    fullInputString
	        else
		    (* continue reading another line of input *)
		    readLine(continuePrompt, fullInputString)
	    end
    in
        readLine(initialPrompt, "")
    end


fun repl() =
    let fun loop(env:Environment) =
        let
            (* read input *)
            val inputString:string = readInput()

	    (* parse the input into an abstract syntax tree *)
	    val ast:TopLevel = parse(inputString)
		   
	    (* print the parsed input *)
	    val _ = print("(* Evaluating " ^ topLevelToString(ast) ^ " *)\n")

	    (* evaluate the parsed input, producing a set of new bindings *)
            val bindings:Environment = eval(ast, env)

	    (* print the result bindings *)
	    val _ = print(envToString(bindings))

	    (* add in the new bindings, shadowing the previous ones *)
	    val new_env:Environment = combine_envs(env, bindings)
	in
	    (* loop *)
	    loop(new_env)
        end

	handle	       
	    (* handle any errors that might have happened in the any of
	       the steps above *)

	    ParseError(msg:string) =>
	        (* there was a syntax error.  print the error message,
		   then continue the repl loop *)
		(print("syntax error: " ^ msg ^ "\n");
		 loop(env))

	  | EvalError(msg:string) =>
	        (* there was an error during evaluation.  print the
		   error message, then continue the repl loop *)
		(print("evaluation error: " ^ msg ^ "\n");
		 loop(env))

	  | ImplementMe(feature:string) =>
	        (* an unhandled case was encounted *)
		(print(feature ^ " not yet implemented\n");
		 loop(env))

    in
       (printWelcome();
        (* start off the repl loop with empty environment *)
        loop(empty_environment))
    end

    handle EmptyInput =>
        (* we're done! *)
	"Thanks for playing.  Please come again soon!\n"

end

end
